@@ -8,26 +8,46 @@ use crate::{
     Data, Hasher, Index, MultiKeychain, NodeIndex, Receiver, Sender, SessionId,
 };
 use codec::{Decode, Encode};
+use derivative::Derivative;
 use futures::{
     channel::{mpsc, oneshot},
     FutureExt, StreamExt,
 };
-use log::{debug, error};
+use log::{debug, error, info, trace, warn};
+use parking_lot::RwLock;
 use std::{
-    cell::RefCell,
     collections::{HashMap, HashSet},
+    ops::Deref,
     time,
 };
 
 pub(crate) type ForkProof<H, D, S> = (UncheckedSignedUnit<H, D, S>, UncheckedSignedUnit<H, D, S>);
 
-#[derive(Clone, Debug, Decode, Encode)]
+#[derive(Debug, Decode, Encode, Derivative)]
+#[derivative(PartialEq, Eq, Hash)]
 pub(crate) struct Alert<H: Hasher, D: Data, S: Signature> {
     sender: NodeIndex,
     proof: ForkProof<H, D, S>,
     legit_units: Vec<UncheckedSignedUnit<H, D, S>>,
     #[codec(skip)]
-    hash: RefCell<Option<H::Hash>>,
+    #[derivative(PartialEq = "ignore")]
+    #[derivative(Hash = "ignore")]
+    hash: RwLock<Option<H::Hash>>,
+}
+
+impl<H: Hasher, D: Data, S: Signature> Clone for Alert<H, D, S> {
+    fn clone(&self) -> Self {
+        let hash = match self.hash.try_read() {
+            None => None,
+            Some(guard) => *guard.deref(),
+        };
+        Alert {
+            sender: self.sender,
+            proof: self.proof.clone(),
+            legit_units: self.legit_units.clone(),
+            hash: RwLock::new(hash),
+        }
+    }
 }
 
 impl<H: Hasher, D: Data, S: Signature> Alert<H, D, S> {
@@ -40,16 +60,16 @@ impl<H: Hasher, D: Data, S: Signature> Alert<H, D, S> {
             sender,
             proof,
             legit_units,
-            hash: RefCell::new(None),
+            hash: RwLock::new(None),
         }
     }
     fn hash(&self) -> H::Hash {
-        let hash = *self.hash.borrow();
+        let hash = *self.hash.read();
         match hash {
             Some(hash) => hash,
             None => {
                 let hash = self.using_encoded(H::hash);
-                *self.hash.borrow_mut() = Some(hash);
+                *self.hash.write() = Some(hash);
                 hash
             }
         }
@@ -60,6 +80,14 @@ impl<H: Hasher, D: Data, S: Signature> Alert<H, D, S> {
     fn forker(&self) -> NodeIndex {
         self.proof.0.as_signable().creator()
     }
+
+    pub(crate) fn included_data(&self) -> Vec<D> {
+        // Only legit units might end up in the DAG, we can ignore the fork proof.
+        self.legit_units
+            .iter()
+            .map(|uu| uu.as_signable().data().clone())
+            .collect()
+    }
 }
 
 impl<H: Hasher, D: Data, S: Signature> Index for Alert<H, D, S> {
@@ -76,7 +104,7 @@ impl<H: Hasher, D: Data, S: Signature> Signable for Alert<H, D, S> {
 }
 
 /// A message concerning alerts.
-#[derive(Debug, Encode, Decode, Clone)]
+#[derive(Debug, Encode, Decode, Clone, PartialEq, Eq, Hash)]
 pub(crate) enum AlertMessage<H: Hasher, D: Data, S: Signature, MS: PartialMultisignature> {
     /// Alert regarding forks, signed by the person claiming misconduct.
     ForkAlert(UncheckedSigned<Alert<H, D, S>, S>),
@@ -86,13 +114,28 @@ pub(crate) enum AlertMessage<H: Hasher, D: Data, S: Signature, MS: PartialMultis
     AlertRequest(NodeIndex, H::Hash),
 }
 
+impl<H: Hasher, D: Data, S: Signature, MS: PartialMultisignature> AlertMessage<H, D, S, MS> {
+    pub(crate) fn included_data(&self) -> Vec<D> {
+        match self {
+            Self::ForkAlert(unchecked_alert) => unchecked_alert.as_signable().included_data(),
+            Self::RmcMessage(_, _) => Vec::new(),
+            Self::AlertRequest(_, _) => Vec::new(),
+        }
+    }
+}
+
 // Notifications being sent to consensus, so that it can learn about proven forkers and receive
 // legitimized units.
+#[derive(Debug, PartialEq, Eq, Hash)]
 pub(crate) enum ForkingNotification<H: Hasher, D: Data, S: Signature> {
     Forker(ForkProof<H, D, S>),
     Units(Vec<UncheckedSignedUnit<H, D, S>>),
 }
 
+/// The component responsible for fork alerts in AlephBFT. We refer to the documentation
+/// https://cardinal-cryptography.github.io/AlephBFT/how_alephbft_does_it.html Section 2.5 and
+/// https://cardinal-cryptography.github.io/AlephBFT/reliable_broadcast.html and to the Aleph
+/// paper https://arxiv.org/abs/1908.05156 Appendix A1 for a discussion.
 struct Alerter<'a, H: Hasher, D: Data, MK: MultiKeychain> {
     session_id: SessionId,
     keychain: &'a MK,
@@ -103,20 +146,16 @@ struct Alerter<'a, H: Hasher, D: Data, MK: MultiKeychain> {
     messages_from_network: Receiver<AlertMessage<H, D, MK::Signature, MK::PartialMultisignature>>,
     notifications_for_units: Sender<ForkingNotification<H, D, MK::Signature>>,
     alerts_from_units: Receiver<Alert<H, D, MK::Signature>>,
-    max_units_per_alert: usize,
     known_forkers: HashMap<NodeIndex, ForkProof<H, D, MK::Signature>>,
     known_alerts: HashMap<H::Hash, Signed<'a, Alert<H, D, MK::Signature>, MK>>,
     known_rmcs: HashMap<(NodeIndex, NodeIndex), H::Hash>,
     rmc: ReliableMulticast<'a, H::Hash, MK>,
-    messages_from_rmc: Receiver<(
-        NodeIndex,
-        rmc::Message<H::Hash, MK::Signature, MK::PartialMultisignature>,
-    )>,
+    messages_from_rmc: Receiver<rmc::Message<H::Hash, MK::Signature, MK::PartialMultisignature>>,
     messages_for_rmc: Sender<rmc::Message<H::Hash, MK::Signature, MK::PartialMultisignature>>,
+    exiting: bool,
 }
 
 pub(crate) struct AlertConfig {
-    pub max_units_per_alert: usize,
     pub n_members: NodeCount,
     pub session_id: SessionId,
 }
@@ -144,7 +183,6 @@ impl<'a, H: Hasher, D: Data, MK: MultiKeychain> Alerter<'a, H, D, MK> {
             messages_from_network,
             notifications_for_units,
             alerts_from_units,
-            max_units_per_alert: config.max_units_per_alert,
             known_forkers: HashMap::new(),
             known_alerts: HashMap::new(),
             known_rmcs: HashMap::new(),
@@ -157,6 +195,7 @@ impl<'a, H: Hasher, D: Data, MK: MultiKeychain> Alerter<'a, H, D, MK> {
             ),
             messages_from_rmc,
             messages_for_rmc,
+            exiting: false,
         }
     }
 
@@ -171,9 +210,7 @@ impl<'a, H: Hasher, D: Data, MK: MultiKeychain> Alerter<'a, H, D, MK> {
     fn on_new_forker_detected(&mut self, forker: NodeIndex, proof: ForkProof<H, D, MK::Signature>) {
         use ForkingNotification::Forker;
         self.known_forkers.insert(forker, proof.clone());
-        if let Err(e) = self.notifications_for_units.unbounded_send(Forker(proof)) {
-            error!(target: "alerter", "{:?} Failed to send for proof to units: {:?}.", self.index(), e);
-        }
+        self.send_notification_for_units(Forker(proof));
     }
 
     // Correctness rules:
@@ -188,26 +225,22 @@ impl<'a, H: Hasher, D: Data, MK: MultiKeychain> Alerter<'a, H, D, MK> {
         forker: NodeIndex,
         units: &[UncheckedSignedUnit<H, D, MK::Signature>],
     ) -> bool {
-        if units.len() > self.max_units_per_alert {
-            debug!(target: "alerter", "{:?} Too many units: {} included in alert.", self.index(), units.len());
-            return false;
-        }
-        let mut rounds: HashSet<usize> = HashSet::new();
+        let mut rounds = HashSet::new();
         for u in units {
             let u = match u.clone().check(self.keychain) {
                 Ok(u) => u,
                 Err(_) => {
-                    debug!(target: "alerter", "{:?} One of the units is incorrectly signed.", self.index());
+                    warn!(target: "AlephBFT-alerter", "{:?} One of the units is incorrectly signed.", self.index());
                     return false;
                 }
             };
             let full_unit = u.as_signable();
             if full_unit.creator() != forker {
-                debug!(target: "alerter", "{:?} One of the units {:?} has wrong creator.", self.index(), full_unit);
+                warn!(target: "AlephBFT-alerter", "{:?} One of the units {:?} has wrong creator.", self.index(), full_unit);
                 return false;
             }
             if rounds.contains(&full_unit.round()) {
-                debug!(target: "alerter", "{:?} Two or more alerted units have the same round {:?}.", self.index(), full_unit.round());
+                warn!(target: "AlephBFT-alerter", "{:?} Two or more alerted units have the same round {:?}.", self.index(), full_unit.round());
                 return false;
             }
             rounds.insert(full_unit.round());
@@ -223,7 +256,7 @@ impl<'a, H: Hasher, D: Data, MK: MultiKeychain> Alerter<'a, H, D, MK> {
             match (u1, u2) {
                 (Ok(u1), Ok(u2)) => (u1, u2),
                 _ => {
-                    debug!(target: "alerter", "{:?} Invalid signatures in a proof.", self.index());
+                    warn!(target: "AlephBFT-alerter", "{:?} Invalid signatures in a proof.", self.index());
                     return None;
                 }
             }
@@ -232,60 +265,62 @@ impl<'a, H: Hasher, D: Data, MK: MultiKeychain> Alerter<'a, H, D, MK> {
         let full_unit2 = u2.as_signable();
         if full_unit1.session_id() != self.session_id || full_unit2.session_id() != self.session_id
         {
-            debug!(target: "alerter", "{:?} Alert from different session.", self.index());
+            warn!(target: "AlephBFT-alerter", "{:?} Alert from different session.", self.index());
             return None;
         }
         if full_unit1 == full_unit2 {
-            debug!(target: "alerter", "{:?} Two copies of the same unit do not constitute a fork.", self.index());
+            warn!(target: "AlephBFT-alerter", "{:?} Two copies of the same unit do not constitute a fork.", self.index());
             return None;
         }
         if full_unit1.creator() != full_unit2.creator() {
-            debug!(target: "alerter", "{:?} One of the units creators in proof does not match.", self.index());
+            warn!(target: "AlephBFT-alerter", "{:?} One of the units creators in proof does not match.", self.index());
             return None;
         }
         if full_unit1.round() != full_unit2.round() {
-            debug!(target: "alerter", "{:?} The rounds in proof's units do not match.", self.index());
+            warn!(target: "AlephBFT-alerter", "{:?} The rounds in proof's units do not match.", self.index());
             return None;
         }
         Some(full_unit1.creator())
     }
 
-    fn rmc_alert(&mut self, forker: NodeIndex, alert: Signed<'a, Alert<H, D, MK::Signature>, MK>) {
+    async fn rmc_alert(
+        &mut self,
+        forker: NodeIndex,
+        alert: Signed<'a, Alert<H, D, MK::Signature>, MK>,
+    ) {
         let hash = alert.as_signable().hash();
         self.known_rmcs
             .insert((alert.as_signable().sender, forker), hash);
         self.known_alerts.insert(hash, alert);
-        self.rmc.start_rmc(hash);
+        self.rmc.start_rmc(hash).await;
     }
 
-    fn on_own_alert(&mut self, alert: Alert<H, D, MK::Signature>) {
+    async fn on_own_alert(&mut self, alert: Alert<H, D, MK::Signature>) {
         let forker = alert.forker();
         self.known_forkers.insert(forker, alert.proof.clone());
-        let alert = Signed::sign(alert, self.keychain);
-        if let Err(e) = self.messages_for_network.unbounded_send((
+        let alert = Signed::sign(alert, self.keychain).await;
+        self.send_message_for_network(
             AlertMessage::ForkAlert(alert.clone().into()),
             Recipient::Everyone,
-        )) {
-            debug!(target: "alerter", "{:?} Error when broadcasting alert {:?}.", self.index(), e);
-        }
-        self.rmc_alert(forker, alert);
+        );
+        self.rmc_alert(forker, alert).await;
     }
 
-    fn on_network_alert(
+    async fn on_network_alert(
         &mut self,
         alert: UncheckedSigned<Alert<H, D, MK::Signature>, MK::Signature>,
     ) {
         let alert = match alert.check(self.keychain) {
             Ok(alert) => alert,
             Err(e) => {
-                debug!(target: "alerter","{:?} We have received an incorrectly signed alert: {:?}.", self.index(), e);
+                warn!(target: "AlephBFT-alerter","{:?} We have received an incorrectly signed alert: {:?}.", self.index(), e);
                 return;
             }
         };
         let contents = alert.as_signable();
         if let Some(forker) = self.who_is_forking(&contents.proof) {
             if self.known_rmcs.contains_key(&(contents.sender, forker)) {
-                debug!(target: "alerter","{:?} We already know about an alert by {:?} about {:?}.", self.index(), alert.as_signable().sender, forker);
+                debug!(target: "AlephBFT-alerter","{:?} We already know about an alert by {:?} about {:?}.", self.index(), alert.as_signable().sender, forker);
                 self.known_alerts.insert(contents.hash(), alert);
                 return;
             }
@@ -293,9 +328,9 @@ impl<'a, H: Hasher, D: Data, MK: MultiKeychain> Alerter<'a, H, D, MK> {
                 // We learn about this forker for the first time, need to send our own alert
                 self.on_new_forker_detected(forker, contents.proof.clone());
             }
-            self.rmc_alert(forker, alert);
+            self.rmc_alert(forker, alert).await;
         } else {
-            debug!(target: "alerter","{:?} We have received an incorrect forking proof from {:?}.", self.index(), alert.as_signable().sender);
+            warn!(target: "AlephBFT-alerter","{:?} We have received an incorrect forking proof from {:?}.", self.index(), alert.as_signable().sender);
         }
     }
 
@@ -303,16 +338,11 @@ impl<'a, H: Hasher, D: Data, MK: MultiKeychain> Alerter<'a, H, D, MK> {
         let alert = match self.known_alerts.get(&hash) {
             Some(alert) => alert.clone(),
             None => {
-                debug!(target: "alerter", "{:?} Received request for unknown alert.", self.index());
+                debug!(target: "AlephBFT-alerter", "{:?} Received request for unknown alert.", self.index());
                 return;
             }
         };
-        if let Err(e) = self
-            .messages_for_network
-            .unbounded_send((AlertMessage::ForkAlert(alert.into()), Recipient::Node(node)))
-        {
-            debug!(target: "alerter", "{:?} Error when sending alert to {:?}: {:?}.", self.index(), node, e);
-        }
+        self.send_message_for_network(AlertMessage::ForkAlert(alert.into()), Recipient::Node(node));
     }
 
     fn on_rmc_message(
@@ -324,27 +354,29 @@ impl<'a, H: Hasher, D: Data, MK: MultiKeychain> Alerter<'a, H, D, MK> {
         if let Some(alert) = self.known_alerts.get(hash) {
             let alert_id = (alert.as_signable().sender, alert.as_signable().forker());
             if self.known_rmcs.get(&alert_id) == Some(hash) || message.is_complete() {
-                if let Err(e) = self.messages_for_rmc.unbounded_send(message) {
-                    debug!(target: "alerter", "{:?} Error when forwarding RMC message to RMC: {:?}.", self.index(), e);
+                let result = self.messages_for_rmc.unbounded_send(message);
+                if result.is_err() {
+                    warn!(target: "AlephBFT-alerter", "{:?} Channel with messages for rmc should be open", self.index());
+                    self.exiting = true;
                 }
             }
-        } else if let Err(e) = self.messages_for_network.unbounded_send((
-            AlertMessage::AlertRequest(self.index(), *hash),
-            Recipient::Node(sender),
-        )) {
-            debug!(target: "alerter", "{:?} Error when sending alert request: {:?}.", self.index(), e);
+        } else {
+            self.send_message_for_network(
+                AlertMessage::AlertRequest(self.index(), *hash),
+                Recipient::Node(sender),
+            );
         }
     }
 
-    fn on_message(
+    async fn on_message(
         &mut self,
         message: AlertMessage<H, D, MK::Signature, MK::PartialMultisignature>,
     ) {
         use AlertMessage::*;
         match message {
             ForkAlert(alert) => {
-                debug!(target: "alerter", "{:?} Fork alert received {:?}.", self.index(), alert);
-                self.on_network_alert(alert);
+                trace!(target: "AlephBFT-alerter", "{:?} Fork alert received {:?}.", self.index(), alert);
+                self.on_network_alert(alert).await;
             }
             RmcMessage(sender, message) => self.on_rmc_message(sender, message),
             AlertRequest(node, hash) => self.send_alert_to(hash, node),
@@ -355,37 +387,56 @@ impl<'a, H: Hasher, D: Data, MK: MultiKeychain> Alerter<'a, H, D, MK> {
         let alert = match self.known_alerts.get(multisigned.as_signable()) {
             Some(alert) => alert.as_signable(),
             None => {
-                error!(target: "alerter", "{:?} Completed an RMC for an unknown alert.", self.index());
+                error!(target: "AlephBFT-alerter", "{:?} Completed an RMC for an unknown alert.", self.index());
                 return;
             }
         };
         let forker = alert.proof.0.as_signable().creator();
         self.known_rmcs.insert((alert.sender, forker), alert.hash());
         if !self.correct_commitment(forker, &alert.legit_units) {
-            debug!(target: "alerter","{:?} We have received an incorrect unit commitment from {:?}.", self.index(), alert.sender);
+            warn!(target: "AlephBFT-alerter","{:?} We have received an incorrect unit commitment from {:?}.", self.index(), alert.sender);
             return;
         }
-        if let Err(e) = self
-            .notifications_for_units
-            .unbounded_send(ForkingNotification::Units(alert.legit_units.clone()))
-        {
-            debug!(target: "alerter","{:?} Failed to send forking notification to units: {:?}.", self.index(), e);
-        }
+        let notification = ForkingNotification::Units(alert.legit_units.clone());
+        self.send_notification_for_units(notification);
     }
 
     fn rmc_message_to_network(
         &mut self,
-        message: (
-            NodeIndex,
-            rmc::Message<H::Hash, MK::Signature, MK::PartialMultisignature>,
-        ),
+        message: rmc::Message<H::Hash, MK::Signature, MK::PartialMultisignature>,
     ) {
-        let (node, message) = message;
-        if let Err(e) = self.messages_for_network.unbounded_send((
+        self.send_message_for_network(
             AlertMessage::RmcMessage(self.index(), message),
-            Recipient::Node(node),
-        )) {
-            debug!(target: "alerter", "{:?} Error when sending RMC message to {:?}: {:?}.", self.index(), node, e);
+            Recipient::Everyone,
+        );
+    }
+
+    fn send_notification_for_units(
+        &mut self,
+        notification: ForkingNotification<H, D, MK::Signature>,
+    ) {
+        if self
+            .notifications_for_units
+            .unbounded_send(notification)
+            .is_err()
+        {
+            warn!(target: "AlephBFT-alerter", "{:?} Channel with forking notifications should be open", self.index());
+            self.exiting = true;
+        }
+    }
+
+    fn send_message_for_network(
+        &mut self,
+        message: AlertMessage<H, D, MK::Signature, MK::PartialMultisignature>,
+        recipient: Recipient,
+    ) {
+        if self
+            .messages_for_network
+            .unbounded_send((message, recipient))
+            .is_err()
+        {
+            warn!(target: "AlephBFT-alerter", "{:?} Channel with notifications for network should be open", self.index());
+            self.exiting = true;
         }
     }
 
@@ -393,28 +444,35 @@ impl<'a, H: Hasher, D: Data, MK: MultiKeychain> Alerter<'a, H, D, MK> {
         loop {
             futures::select! {
                 message = self.messages_from_network.next() => match message {
-                    Some(message) => self.on_message(message),
+                    Some(message) => self.on_message(message).await,
                     None => {
-                        error!(target: "alerter", "{:?} Message stream closed.", self.index());
+                        error!(target: "AlephBFT-alerter", "{:?} Message stream closed.", self.index());
                         break;
                     }
                 },
                 alert = self.alerts_from_units.next() => match alert {
-                    Some(alert) => self.on_own_alert(alert),
+                    Some(alert) => self.on_own_alert(alert).await,
                     None => {
-                        error!(target: "alerter", "{:?} Alert stream closed.", self.index());
+                        error!(target: "AlephBFT-alerter", "{:?} Alert stream closed.", self.index());
                         break;
                     }
                 },
                 message = self.messages_from_rmc.next() => match message {
                     Some(message) => self.rmc_message_to_network(message),
                     None => {
-                        error!(target: "alerter", "{:?} RMC message stream closed.", self.index());
+                        error!(target: "AlephBFT-alerter", "{:?} RMC message stream closed.", self.index());
                         break;
                     }
                 },
                 multisigned = self.rmc.next_multisigned_hash().fuse() => self.alert_confirmed(multisigned),
-                _ = &mut exit => break,
+                _ = &mut exit => {
+                    info!(target: "AlephBFT-alerter", "{:?} received exit signal", self.index());
+                    self.exiting = true;
+                },
+            }
+            if self.exiting {
+                info!(target: "AlephBFT-alerter", "{:?} Alerter decided to exit.", self.index());
+                break;
             }
         }
     }
