@@ -942,6 +942,7 @@ pub(crate) async fn run<H, D, MK, DP, FH, SH>(
         if let Err(()) = consensus_handle.await {
             warn!(target: "AlephBFT-runway", "{:?} Consensus finished with an error", index);
         }
+        debug!(target: "AlephBFT-runway", "{:?} Consensus stopped.", index);
     }
 
     if alerter_exit.send(()).is_err() {
@@ -951,6 +952,7 @@ pub(crate) async fn run<H, D, MK, DP, FH, SH>(
         if let Err(()) = alerter_handle.await {
             warn!(target: "AlephBFT-runway", "{:?} Alerter finished with an error", index);
         }
+        debug!(target: "AlephBFT-runway", "{:?} Alerter stopped.", index);
     }
 
     if runway_exit.send(()).is_err() {
@@ -958,6 +960,7 @@ pub(crate) async fn run<H, D, MK, DP, FH, SH>(
     }
     if !runway_handle.is_terminated() {
         runway_handle.await;
+        debug!(target: "AlephBFT-runway", "{:?} Runway stopped.", index);
     }
 
     info!(target: "AlephBFT-runway", "{:?} Runway ended.", index);
