@@ -3,12 +3,12 @@ use std::collections::{hash_map::Entry, HashMap, VecDeque};
 
 use crate::{
     extender::ExtenderUnit,
-    member::{NotificationIn, NotificationOut},
     nodes::{NodeCount, NodeIndex, NodeMap},
+    runway::{NotificationIn, NotificationOut},
     units::{ControlHash, Unit, UnitCoord},
     Hasher, Receiver, Round, Sender,
 };
-use log::{debug, error};
+use log::{debug, info, trace, warn};
 
 /// An enum describing the status of a Unit in the Terminal pipeline.
 #[derive(Clone, PartialEq)]
@@ -28,7 +28,7 @@ pub struct TerminalUnit<H: Hasher> {
     // and hashes of parents are being added at the time we receive a given coord. Once the parents map is complete
     // we test whether the hash of parents agains the control_hash in the unit. If the check fails, we request hashes
     // of parents of this unit via a NotificationOut.
-    parents: NodeMap<Option<H::Hash>>,
+    parents: NodeMap<H::Hash>,
     n_miss_par_decoded: NodeCount,
     n_miss_par_dag: NodeCount,
     status: UnitStatus,
@@ -53,7 +53,7 @@ impl<H: Hasher> TerminalUnit<H> {
         let n_parents = unit.control_hash().n_parents();
         TerminalUnit {
             unit: unit.clone(),
-            parents: NodeMap::new_with_len(n_members),
+            parents: NodeMap::with_size(n_members),
             n_miss_par_decoded: n_parents,
             n_miss_par_dag: n_parents,
             status: UnitStatus::ReconstructingParents,
@@ -98,6 +98,9 @@ type SyncClosure<X, Y> = Box<dyn Fn(X) -> Y + Sync + Send + 'static>;
 ///    gets this status, we mark all the parents that are already in the Dag and set appropriate triggers for when
 ///    the remaining parents are added to Dag.
 /// 6) At the moment when all parents have been added to Dag, the unit itself is added to Dag.
+///
+/// We also refer to the documentation https://cardinal-cryptography.github.io/AlephBFT/internals.html
+/// Section 5.3 for a discussion of this component.
 
 pub(crate) struct Terminal<H: Hasher> {
     node_id: NodeIndex,
@@ -122,6 +125,7 @@ pub(crate) struct Terminal<H: Hasher> {
     // The same as above, but this time we await for a unit (with a particular hash) to be added to the Dag.
     // Once this happens, we notify all the children.
     children_hash: HashMap<H::Hash, Vec<H::Hash>>,
+    exiting: bool,
 }
 
 impl<H: Hasher> Terminal<H> {
@@ -140,6 +144,7 @@ impl<H: Hasher> Terminal<H> {
             unit_by_coord: HashMap::new(),
             children_coord: HashMap::new(),
             children_hash: HashMap::new(),
+            exiting: false,
         }
     }
 
@@ -148,7 +153,7 @@ impl<H: Hasher> Terminal<H> {
         let u = self.unit_store.get_mut(u_hash).unwrap();
         // the above unwraps must succeed, should probably add some debug messages here...
 
-        u.parents[pid] = Some(*p_hash);
+        u.parents.insert(pid, *p_hash);
         u.n_miss_par_decoded -= NodeCount(1);
         if u.n_miss_par_decoded == NodeCount(0) {
             self.event_queue
@@ -191,9 +196,14 @@ impl<H: Hasher> Terminal<H> {
     fn update_on_store_add(&mut self, u: Unit<H>) {
         let u_hash = u.hash();
         let (u_round, pid) = (u.round(), u.creator());
-        // If u is a fork, then the below line will overwrite the previous unit at this coord, but this is intended
-        // and does not break correctness.
-        self.unit_by_coord.insert((u_round, pid), u_hash);
+        // We place the unit in the coord map only if this is the first variant ever received.
+        // This is not crucial for correctness, but helps in clarity.
+        if let Entry::Vacant(entry) = self.unit_by_coord.entry((u_round, pid)) {
+            entry.insert(u_hash);
+        } else {
+            debug!(target: "AlephBFT-terminal", "Received a fork at round {:?} creator {:?}", u_round, pid);
+        }
+
         if let Some(children) = self.children_coord.remove(&(u_round, pid)) {
             // We reconstruct the parent for each unit that waits for this coord.
             for v_hash in children {
@@ -218,13 +228,8 @@ impl<H: Hasher> Terminal<H> {
                 }
             }
             if !coords_to_request.is_empty() {
-                debug!(target: "aleph-terminal", "{:?} Missing coords {:?}", self.node_id, coords_to_request);
-                let send_result = self
-                    .ntfct_tx
-                    .unbounded_send(NotificationOut::MissingUnits(coords_to_request));
-                if let Err(e) = send_result {
-                    error!(target: "aleph-terminal", "{:?} Unable to place a Fetch request: {:?}.", self.node_id, e);
-                }
+                trace!(target: "AlephBFT-terminal", "{:?} Missing coords {:?}", self.node_id, coords_to_request);
+                self.send_notification(NotificationOut::MissingUnits(coords_to_request));
             }
         }
     }
@@ -242,16 +247,11 @@ impl<H: Hasher> Terminal<H> {
             }
         }
         let mut parent_hashes = Vec::new();
-        for p_hash in u.parents.iter().flatten() {
-            parent_hashes.push(*p_hash);
+        for p_hash in u.parents.into_values() {
+            parent_hashes.push(p_hash);
         }
 
-        let send_result = self
-            .ntfct_tx
-            .unbounded_send(NotificationOut::AddedToDag(*u_hash, parent_hashes));
-        if let Err(e) = send_result {
-            error!(target: "aleph-terminal", "{:?} Unable to place AddedToDag notification: {:?}.", self.node_id, e);
-        }
+        self.send_notification(NotificationOut::AddedToDag(*u_hash, parent_hashes));
     }
 
     // We set the correct parent hashes for unit u.
@@ -261,19 +261,19 @@ impl<H: Hasher> Terminal<H> {
             .get_mut(&u_hash)
             .expect("unit with wrong control hash must be in store");
         if u.status != UnitStatus::WrongControlHash {
-            debug!(target: "aleph-terminal", "{:?} Received parents response without it being expected for {:?}. Ignoring.", self.node_id, u_hash);
+            trace!(target: "AlephBFT-terminal", "{:?} Received parents response without it being expected for {:?}. Ignoring.", self.node_id, u_hash);
             return;
         }
         for (counter, i) in u.unit.control_hash().parents().enumerate() {
-            u.parents[i] = Some(p_hashes[counter]);
+            u.parents.insert(i, p_hashes[counter]);
         }
-        debug!(target: "aleph-terminal", "{:?} Updating parent hashes for wrong control hash unit {:?}", self.node_id, u_hash);
+        trace!(target: "AlephBFT-terminal", "{:?} Updating parent hashes for wrong control hash unit {:?}", self.node_id, u_hash);
         u.n_miss_par_decoded = NodeCount(0);
         self.inspect_parents_in_dag(&u_hash);
     }
 
     fn add_to_store(&mut self, u: Unit<H>) {
-        debug!(target: "aleph-terminal", "{:?} Adding to store {:?} round {:?} index {:?}", self.node_id, u.hash(), u.round(), u.creator());
+        trace!(target: "AlephBFT-terminal", "{:?} Adding to store {:?} round {:?} index {:?}", self.node_id, u.hash(), u.round(), u.creator());
         if let Entry::Vacant(entry) = self.unit_store.entry(u.hash()) {
             entry.insert(TerminalUnit::<H>::blank_from_unit(&u));
             self.update_on_store_add(u);
@@ -281,9 +281,9 @@ impl<H: Hasher> Terminal<H> {
     }
 
     fn inspect_parents_in_dag(&mut self, u_hash: &H::Hash) {
-        let u_parents = self.unit_store.get(&u_hash).unwrap().parents.clone();
+        let u_parents = self.unit_store.get(u_hash).unwrap().parents.clone();
         let mut n_parents_in_dag = NodeCount(0);
-        for p_hash in u_parents.into_iter().flatten() {
+        for p_hash in u_parents.into_values() {
             let maybe_p = self.unit_store.get(&p_hash);
             // p might not be even in store because u might be a unit with wrong control hash
             match maybe_p {
@@ -295,9 +295,9 @@ impl<H: Hasher> Terminal<H> {
                 }
             }
         }
-        let u = self.unit_store.get_mut(&u_hash).unwrap();
+        let u = self.unit_store.get_mut(u_hash).unwrap();
         u.n_miss_par_dag -= n_parents_in_dag;
-        debug!(target: "aleph-terminal", "{:?} Inspecting parents for {:?}, missing {:?}", self.node_id, u_hash, u.n_miss_par_dag);
+        trace!(target: "AlephBFT-terminal", "{:?} Inspecting parents for {:?}, missing {:?}", self.node_id, u_hash, u.n_miss_par_dag);
         if u.n_miss_par_dag == NodeCount(0) {
             self.event_queue
                 .push_back(TerminalEvent::ParentsInDag(*u_hash));
@@ -308,12 +308,7 @@ impl<H: Hasher> Terminal<H> {
     }
 
     fn on_wrong_hash_detected(&mut self, u_hash: H::Hash) {
-        let send_result = self
-            .ntfct_tx
-            .unbounded_send(NotificationOut::WrongControlHash(u_hash));
-        if let Err(e) = send_result {
-            error!(target: "aleph-terminal", "{:?} Unable to place a Fetch request: {:?}.", self.node_id, e);
-        }
+        self.send_notification(NotificationOut::WrongControlHash(u_hash));
     }
 
     // This drains the event queue. Note that new events might be added to the queue as the result of
@@ -329,14 +324,14 @@ impl<H: Hasher> Terminal<H> {
                         self.inspect_parents_in_dag(&u_hash);
                     } else {
                         u.status = UnitStatus::WrongControlHash;
-                        debug!(target: "aleph-terminal", "{:?} wrong control hash", self.node_id);
+                        warn!(target: "AlephBFT-terminal", "{:?} wrong control hash", self.node_id);
                         self.on_wrong_hash_detected(u_hash);
                     }
                 }
                 TerminalEvent::ParentsInDag(u_hash) => {
                     let u = self.unit_store.get_mut(&u_hash).unwrap();
                     u.status = UnitStatus::InDag;
-                    debug!(target: "aleph-terminal", "{:?} Adding to Dag {:?} round {:?} index {:?}.", self.node_id, u_hash, u.unit.round(), u.unit.creator());
+                    trace!(target: "AlephBFT-terminal", "{:?} Adding to Dag {:?} round {:?} index {:?}.", self.node_id, u_hash, u.unit.round(), u.unit.creator());
                     self.update_on_dag_add(&u_hash);
                 }
             }
@@ -347,6 +342,13 @@ impl<H: Hasher> Terminal<H> {
         self.post_insert.push(hook);
     }
 
+    fn send_notification(&mut self, notification: NotificationOut<H>) {
+        if self.ntfct_tx.unbounded_send(notification).is_err() {
+            warn!(target: "AlephBFT-terminal", "{:?} Notification channel should be open", self.node_id);
+            self.exiting = true;
+        }
+    }
+
     pub(crate) async fn run(&mut self, mut exit: oneshot::Receiver<()>) {
         loop {
             futures::select! {
@@ -366,10 +368,14 @@ impl<H: Hasher> Terminal<H> {
                     }
                 }
                 _ = &mut exit => {
-                    debug!(target: "aleph-terminal", "{:?} received exit signal.", self.node_id);
-                    break
+                    info!(target: "AlephBFT-terminal", "{:?} received exit signal", self.node_id);
+                    self.exiting = true;
                 }
             }
+            if self.exiting {
+                info!(target: "AlephBFT-terminal", "{:?} Terminal decided to exit.", self.node_id);
+                break;
+            }
         }
     }
 }
